#include <iostream>
#include <emmintrin.h>
#include <conio.h>
#include <string.h>
#include <time.h>
#include <vector>
#include <iomanip>
#include <assert.h>
#include <fstream>

#ifndef null 
#define null 0
#endif

// str_find 
// we want &str[i] : str[i] == c, i > 0, i < str.length or null

const char* str_find_char_0 (const char* buffer, size_t length, char c) 
{
	const char* end(buffer + length);
	for (const char* p(buffer); p < end; ++p)
		if (*p == c)
			return p;

	return null;
}

const __int32 lowest_bit_set_lookup[16] = { 0, 1, 2, 1, 3, 1, 2, 1, 4, 1, 2, 1, 3, 1, 2, 1 };

// mask is 16 bit result of _mm_movemask_epi8
inline __int32 lowest_bit_set (__int32 mask)
{
	if (mask & 0x00ff) {
		if (mask & 0x000f) {
			return lowest_bit_set_lookup[mask & 0xf] - 1;
		} else {
			return lowest_bit_set_lookup[(mask>>4) & 0xf] + 3;
		}
	} else {
		if (mask & 0x0f00) {
			return lowest_bit_set_lookup[(mask>>8) & 0xf] + 7;
		} else {
			return lowest_bit_set_lookup[(mask>>12) & 0xf] + 11;
		}
	}
}

const __int32 highest_bit_set_lookup[16] = { 0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4 };

// mask is 16 bit result of _mm_movemask_epi8
inline __int32 highest_bit_set (__int32 mask)
{
	if (mask & 0xff00) {
		if (mask & 0xf000) {
			return highest_bit_set_lookup[(mask>>12) & 0xf] + 11;
		} else {
			return highest_bit_set_lookup[(mask>>8) & 0xf] + 7;
		}
	} else {
		if (mask & 0x00f0) {
			return highest_bit_set_lookup[(mask>>4) & 0xf] + 3;
		} else {
			return highest_bit_set_lookup[mask & 0xf] - 1;
		}
	}
}

const char* str_find_char_0x (const char* buffer, size_t length, char c)
{
	__m128i c16(_mm_set1_epi8(c)); // 16 repeats of c

	for (size_t i(0); i < length; i += 16)
	{
		__m128i	test(_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + i)));
		__m128i	compare(_mm_cmpeq_epi8(c16, test));
		__int32	mask(_mm_movemask_epi8(compare));

		if (mask) {
			// add the index of the least significant set bit in mask to i
			i += lowest_bit_set(mask);
			return i < length ? (buffer + i) : null;
		}
	}

	return null;
}

inline __int32 hash32 (__int32 c)
{
	// Bob Jenkin's mix function
	__int32 a,b;
	a = b = 0x7a915b7e;
	a -= b; a -= c; a ^= (c>>13);
	b -= c; b -= a; b ^= (a<<8);
	c -= a; c -= b; c ^= (b>>13);
	a -= b; a -= c; a ^= (c>>12);
	b -= c; b -= a; b ^= (a<<16);
	c -= a; c -= b; c ^= (b>>5);
	a -= b; a -= c; a ^= (c>>3);
	b -= c; b -= a; b ^= (a<<10);
	c -= a; c -= b; c ^= (b>>15);
	return c;
}


// str_find_str_0: naive approach
const char* str_find_str_0 (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length)
{
	const char	first_char(pattern[0]);
	const char*	end(buffer + buffer_length - pattern_length);

	for (const char* p(buffer); p < end; ++p)
		if (*p == first_char) {
			bool match(true);
			for (size_t i(1); i < pattern_length; ++i) {
				if (p[i] != pattern[i]) {
					match = false;
					break;
				}
			}

			if (match) return p;
		}

		return null;
}
// str_find_str_0x: SSE2 adapted naive algorithm
const char* str_find_str_0x (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length) {
	const char	first_char(pattern[0]);

	__int32	final_mask((1 << __int32(pattern_length%16 == 0 ? 16 : (pattern_length%16))) - 1);
	__m128i c16(_mm_set1_epi8(first_char)); // 16 repeats of c
	__m128i	pattern128(_mm_loadu_si128(reinterpret_cast<const __m128i*>(pattern)));

	if (pattern_length > 16) {
		for (size_t i(0); i < buffer_length; i += 16) {
			__m128i	test(_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + i)));
			__m128i	compare(_mm_cmpeq_epi8(c16, test));
			__int32	mask(_mm_movemask_epi8(compare));

			if (mask) {
				// add the index of the least significant set bit in mask to i
				size_t j(i);

				while (mask) {
					__int32 lbs(lowest_bit_set(mask));

					j += lbs;

					if (j >= buffer_length) return null;

					// compare with pattern
					bool matched(true);
					size_t k(0);

					while (matched && ((k+16) < pattern_length)) {
						matched &= (_mm_movemask_epi8(_mm_cmpeq_epi8(
							_mm_loadu_si128(reinterpret_cast<const __m128i*>(pattern + k)), 
							_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + j + k)))) == 0xffff);
						k += 16;
					}

					if (matched && ((_mm_movemask_epi8(_mm_cmpeq_epi8(
						_mm_loadu_si128(reinterpret_cast<const __m128i*>(pattern + k)),  
						_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + j + k)))) & final_mask) == 
						final_mask)) return (buffer + j);

					mask >>= (lbs + 1);
					++j;
				}
			}
		}
	} else {
		for (size_t i(0); i < buffer_length; i += 16)
		{
			__m128i	test(_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + i)));
			__m128i	compare(_mm_cmpeq_epi8(c16, test));
			__int32	mask(_mm_movemask_epi8(compare));

			if (mask) {
				// add the index of the least significant set bit in mask to i
				size_t j(i);

				while (mask) {
					__int32 lbs(lowest_bit_set(mask));

					j += lbs;

					if (j >= buffer_length) return null;

					// compare with pattern
					if ((_mm_movemask_epi8(_mm_cmpeq_epi8(
						pattern128, 
						_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + j)))) & final_mask) == 
						final_mask) return (buffer + j);
	
					mask >>= (lbs + 1);
					++j;
				}
			}
		}
	}

	return null;
}

void construct_offset_table (size_t* offset_table, const char* pattern, size_t pattern_length) {
	for (size_t n(0); n < pattern_length; ++n) {
		const size_t	n_0(pattern_length - 1 - n);
		const char		mismatched_char(pattern[n_0]);
		const size_t	i(pattern_length-n-1);

		offset_table[i] = pattern_length;

		for (size_t shift(1); shift < pattern_length; ++shift) {
			const size_t shift_0(n_0 - shift);

			if ((n_0 < shift) || (pattern[shift_0] != mismatched_char)) {
				const size_t start( (n_0 >= shift) ? 1 : (shift - n_0));
				// attempt to match remainder
				bool matches_coda(true);

				for (size_t j(start); j <= n; ++j) {
					if (pattern[shift_0 + j] != pattern[n_0 + j]) {
						matches_coda = false;
						break;
					}
				}

				if (matches_coda) {
					offset_table[i] = shift;
					break;
				}
			}
		}
	}
}

template <typename T> inline T max (T a, T b) { return a > b ? a : b; }

// str_find_str_1: Boyer-Moore algorithm
const char* str_find_str_1 (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length) 
{
	// construct occurance_table
	__int32	occurance_table[256];

	for (size_t i(0); i < 256; ++i) {
		occurance_table[i] = -1;
	}

	for (size_t i(0); i < pattern_length-1; ++i) {
		occurance_table[pattern[i]] = __int32(i);
	}

	// construct offset table
	size_t*	offset_table = new size_t[pattern_length];

	construct_offset_table(offset_table, pattern, pattern_length);

	// search
	const char* result(null);
	size_t max_pos(buffer_length-pattern_length);

	for (size_t i(0); i <= max_pos; ) {
		size_t j(pattern_length-1);

		while (pattern[j] == buffer[j+i]) {
			if (j == 0) {
				result = buffer + i;
				break;
			}

			--j;
		}

		if (result) break;

		i += max(__int32(offset_table[j]), __int32(j) - occurance_table[buffer[j+i] & 0xff]);
	}

	// clean up
	delete [] offset_table;

	return result;
}

// str_find_str_1x: SSE2 accelerated Boyer-Moore algorithm
const char* str_find_str_1x (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length) 
{
	// haven't done SSE version for > 16 chars
	if (pattern_length > 16) return str_find_str_1(buffer, buffer_length, pattern, pattern_length);

	// construct occurance_table
	__int32	occurance_table[256];

	for (size_t i(0); i < 256; ++i) {
		occurance_table[i] = -1;
	}

	for (size_t i(0); i < pattern_length-1; ++i) {
		occurance_table[pattern[i]] = __int32(i);
	}

	// construct offset table
	size_t*	offset_table = new size_t[pattern_length];

	construct_offset_table(offset_table, pattern, pattern_length);

	// search
	const char* result(null);
	size_t max_pos(buffer_length-pattern_length);

	__m128i	pattern128(_mm_loadu_si128(reinterpret_cast<const __m128i*>(pattern)));
	__int32	final_mask((1 << __int32(pattern_length%16 == 0 ? 16 : (pattern_length%16))) - 1);

	for (size_t i(0); i <= max_pos; ) {
		// compare with pattern
		size_t	j(pattern_length - 1);

		if (pattern[j] == buffer[j+i]) {
				__int32 mask(_mm_movemask_epi8(_mm_cmpeq_epi8(
					pattern128, 
					_mm_loadu_si128(reinterpret_cast<const __m128i*>(buffer + i)))) & final_mask);

				if (mask) {
					if (mask == final_mask) {
						result = (buffer + i);
						break;
					}

					j = highest_bit_set((~mask)&final_mask);
				}
		}
#ifdef _DEBUG
		{
			size_t j2(pattern_length-1);

			while (pattern[j2] == buffer[j2+i]) {
				if (j2 == 0) {
					result = buffer + i;
					break;
				}

				--j2;
			}

			assert(j2 == j);
		}
#endif

		i += max(__int32(offset_table[j]), __int32(j) - occurance_table[buffer[j+i] & 0xff]);
	}

	// clean up
	delete [] offset_table;

	return result;
}

class bmh_matcher
{
public:
	bmh_matcher (const char* pattern, size_t pattern_length) : _pattern(pattern), _pattern_length(pattern_length)
	{
		// construct occurance_table
		__int32*	occurance_table(_occurance_table);

		for (size_t i(0); i < 256; ++i) {
			occurance_table[i] = __int32(pattern_length);
		}

		for (size_t i(0); i < pattern_length; ++i) {
			occurance_table[pattern[i]] = __int32(pattern_length - 1 - i);
		}
	}

private:
	const char*	_pattern;
	size_t		_pattern_length;
	__int32	_occurance_table[256];
};

// str_find_str_2: BoyerMooreHorspool algorithm
const char* str_find_str_2 (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length) 
{
	// search
	for (size_t i(pattern_length-1); i <= buffer_length; ) {
		size_t j(pattern_length-1);

		while (pattern[j] == buffer[i]) {
			if (j == 0) {
				return buffer + i;
			}
			--i;
			--j;
		}

		i += max(__int32(pattern_length - j), _occurance_table[buffer[i] & 0xff]);
	}

	return null;
}

// str_find_str_3: Knuth-Morris-Pratt
// adapted from http://www.ics.uci.edu/~eppstein/161/kmp/

// Make jump table for mismatches (the usual finite state automaton).
//
// The inner loop works by checking the prefixes the next character of
// the target can continue, longest first, chaining back by way of the
// previously filled retarget entries, until it either finds one that
// matches (the != clause) or runs out of prefixes (the > 0 clause).

static int * make_retarget(const char* pattern, size_t pattern_length)
{
    if (pattern == 0) return 0;
    int * retarget = new int[pattern_length+1];
    if (retarget == 0) return 0;
    retarget[0] = -1;		// set up for loop below; unused by match()

	for (size_t i(0); i < pattern_length; ++i) {
		retarget[i + 1] = retarget[i] + 1;
		while (retarget[i + 1] > 0 &&
			pattern[i] != pattern[retarget[i + 1] - 1])
			retarget[i + 1] = retarget[retarget[i + 1] - 1] + 1;
	}
    return retarget;
}

const char* str_find_str_3 (const char* buffer, size_t buffer_length, const char* pattern, size_t pattern_length) 
{
    int*	retarget = make_retarget(pattern, pattern_length);

	if (retarget == 0) return null;

	// Do the search
	size_t index(0);

	for (size_t	i(0); i < buffer_length; ++i) {
		char c(buffer[i]);

		if (c != pattern[index]) {
			do {
				if (index == 0) break; // fell all the way back, have to give up
				index = retarget[index]; // more positions to fall back to, keep trying
			} while (c != pattern[index]);
		} else {
			++index;
			if (index >= pattern_length) {
				//	index = retarget[index]; // full match, but keep going
				delete [] retarget;
				return buffer + i - pattern_length + 1;
			}
		}
	}

	delete [] retarget;
	return null;
}

// the tests
size_t thousand(1000);
size_t billion(thousand * thousand * thousand);
size_t count(1);

void output_result (const char* result, const char* buffer) 
{
	if (result) {
		std::cout << "result = " << std::setbase(10) << size_t(result - buffer) << " <";
		for (size_t i(0); (i < 16) && result[i]; ++i) 
			std::cout << result[i];
		std::cout << ">" << std::endl;
	} else {
		std::cout << "result = null" << std::endl;
	}
}

#define time_find_str(f) \
{\
	clock_t start(clock());\
	\
	size_t		hits(0);\
	\
{\
	for (size_t i(0); i < count; ++i) {\
	for (const char* result(buffer); result = f(result, length + buffer - result, pattern, pattern_length); ++result) \
	++hits; \
	}\
}\
	\
	clock_t finish(clock());\
	std::cout << "pattern found " << hits << " times." << std::endl;\
	std::cout << #f " duration " << (1000 * (double)(finish - start) / CLOCKS_PER_SEC) << "ms" << std::endl;\
}\

#define time_find_char(f) \
{\
	clock_t start(clock());\
	\
	const char* result = null;\
	\
{\
	for (size_t i(0); i < count; ++i) {\
	result = buffer;\
	while(result = f(result, length + buffer - result, (char)(i&0x3f + ' '))) ++result;\
	}\
}\
	\
	clock_t finish(clock());\
	output_result(result, buffer);\
	std::cout << #f " duration " << (1000 * (double)(finish - start) / CLOCKS_PER_SEC) << "ms" << std::endl;\
}\

int main (int argc, char* argv[]) 
{	
	std::cout << std::setprecision(4);

	if (false) {
		char* pattern = "ANPANMAN\0\0\0\0\0\0\0\0\0\0\0\0\0";
		size_t pattern_length(strlen(pattern));

		std::cout << std::setbase(10);
		std::cout << "pattern_length " << pattern_length << std::endl << std::endl;

		size_t*	offset_table = new size_t[pattern_length];

		construct_offset_table(offset_table, pattern, pattern_length);

		for (size_t i(0); i < pattern_length; ++i) std::cout << i << " -> " << offset_table[i] << std::endl;

		std::cout << std::endl;

		delete [] offset_table;
	}

	// nb: lengths must be at least 15 less than allocated memory for safety
	if (false) {
#ifdef _DEBUG
		count = ( 100 * thousand);
#else
		count = ( 5 * thousand * thousand);
#endif
		char buffer[1024];
		size_t length(1000);

		std::cout << std::setbase(10);
		std::cout << "str_find(char)" << std::endl;
		std::cout << "iterations " << count << std::endl;
		std::cout << "length     " << length << std::endl << std::endl;

		for (size_t i(0); i < length; ++i) {
			buffer[i] = ' ' + (hash32(__int32(i)) & 0x2f); // nb: not all chars in test range present
		}

		// check they are functionally the same
		for (size_t i(0); i < count; ++i) {
			const char* result_0  = str_find_char_0(buffer, length, (char)(i&0x3f + ' '));
			const char* result_0x = str_find_char_0x(buffer, length, (char)(i&0x3f + ' '));

			if (result_0 != result_0x) {
				std::cout << "inconsistancy with i = " << i << std::endl << std::endl;
				result_0  = str_find_char_0(buffer, length, (char)(i&0x3f + ' '));
				result_0x = str_find_char_0x(buffer, length, (char)(i&0x3f + ' '));
				std::cout << "  str_find_0  gives "; output_result(result_0, buffer);
				std::cout << "  str_find_0x gives "; output_result(result_0x, buffer);
				break;
			}
		}

		// time them
		for (size_t c(0); c < 7; ++c) {
			time_find_char(str_find_char_0);
			time_find_char(str_find_char_0x);
			std::cout << std::endl;
		}
	}

	{
#ifdef _DEBUG
		count = ( 100 );
#else
		count = ( 10 * thousand);
#endif

		char buffer[1000024];
		size_t length(0);

		std::ifstream	in;
		
		in.open("C:\\kin\\kincore\\richardIII.txt");
		
		while (!in.eof()) {
			in.read(buffer + length, 1000000 - length);
			size_t count(in.gcount());
			if (count == 0) break;
			length += count;
		}

		in.close();

		std::cout << std::setbase(10);
		std::cout << "str_find(string)" << std::endl;
		std::cout << "iterations     " << count << std::endl;
		std::cout << "length         " << length << std::endl << std::endl;

		char* patterns[8] = {
			"discontent",
			"York",
			"horse",
			" she is lunatic.",
			"Would they were basilisks, to strike thee dead!",
			"    I would they were, that I might die at once;\n"\
			"    For now they kill me with a living death.\n"\
			"    Those eyes of thine from mine have drawn salt tears,\n"\
			"    Shamed their aspect with store of childish drops:",
			"I would they were, that I might die at once;\n"\
			"    For now they kill me with a living death.\n"\
			"    Those eyes of thine from mine have drawn salt tears,\n"\
			"    Shamed their aspect with store of childish drops:",
			"________________"
		};

		for (size_t pattern_index(0); pattern_index < 8; ++pattern_index) {
			const char* pattern(patterns[pattern_index]);
			size_t pattern_length(strlen(pattern));

			std::cout << "pattern        <" << pattern << ">" << std::endl;
			std::cout << "pattern_length " << pattern_length << std::endl << std::endl;

			// check they are functionally the same
			for (size_t i(0); i < count; ++i) {
				size_t offset(i % length/2);

				const char* result_0 = str_find_str_0(buffer + offset, length - offset, pattern, pattern_length);
				const char* result_0x = str_find_str_0x(buffer + offset, length - offset, pattern, pattern_length);
				const char* result_1 = str_find_str_1(buffer + offset, length - offset, pattern, pattern_length);
				const char* result_1x = str_find_str_1x(buffer + offset, length - offset, pattern, pattern_length);
				const char* result_2 = str_find_str_2(buffer + offset, length - offset, pattern, pattern_length);
//				const char* result_2x = str_find_str_2x(buffer + offset, length - offset, pattern, pattern_length);
				const char* result_3 = str_find_str_3(buffer + offset, length - offset, pattern, pattern_length);

				if ((result_0 != result_0x) || 
					(result_0 != result_1) || (result_0 != result_1x) || 
					(result_0 != result_2) || 
					(result_0 != result_3) ) {
					std::cout << "inconsistancy with i = " << i << std::endl << std::endl;
					result_0 = str_find_str_0(buffer + offset, length - offset, pattern, pattern_length);
					result_0x = str_find_str_0x(buffer + offset, length - offset, pattern, pattern_length);
					result_1 = str_find_str_1(buffer + offset, length - offset, pattern, pattern_length);
					result_1x = str_find_str_1x(buffer + offset, length - offset, pattern, pattern_length);
					result_2 = str_find_str_2(buffer + offset, length - offset, pattern, pattern_length);
					result_3 = str_find_str_3(buffer + offset, length - offset, pattern, pattern_length);
					std::cout << "  str_find_0  gives "; output_result(result_0, buffer);
					std::cout << "  str_find_0x gives "; output_result(result_0x, buffer);
					std::cout << "  str_find_1  gives "; output_result(result_1, buffer);
					std::cout << "  str_find_1x gives "; output_result(result_1x, buffer);
					std::cout << "  str_find_2  gives "; output_result(result_2, buffer);
					std::cout << "  str_find_3  gives "; output_result(result_3, buffer);
					break;
				}
			}

			// time them
			time_find_str(str_find_str_0);
			time_find_str(str_find_str_0x);
			time_find_str(str_find_str_1);
			time_find_str(str_find_str_1x);
			time_find_str(str_find_str_2);
			time_find_str(str_find_str_3);
			std::cout << std::endl;
			time_find_str(str_find_str_0);
			time_find_str(str_find_str_0x);
			time_find_str(str_find_str_1);
			time_find_str(str_find_str_1x);
			time_find_str(str_find_str_2);
			time_find_str(str_find_str_3);
			std::cout << std::endl;
		}
	}

	// wait for user
	std::cout << "Press any key " << std::endl;
	_getch();

	return 0;
}

